import logging

from discord.ext import commands


def getFullname(o):
    moduleName = o.__class__.__module__
    if moduleName == "builtins":
        moduleName = ""
    if moduleName:
        moduleName = f"{moduleName}."

    className = o.__class__.__name__
    fullname = f"{moduleName}{className}"
    return fullname


class FancyException(Exception):
    level = logging.WARNING

    def formatMessage(self):
        return None

    def formatUserMessage(self):
        return None

    def __repr__(self):
        return getFullname(self)

    def __str__(self):
        return self.formatMessage() or self.formatUserMessage() or repr(self)


class FancyContextException(Exception):
    level = logging.WARNING

    async def formatMessage(self, ctx):
        return None

    async def formatUserMessage(self, ctx):
        return None

    def __repr__(self):
        return getFullname(self)

    def __str__(self):
        return repr(self)


class ArgumentException(FancyContextException):
    async def formatUserMessage(self, ctx):
        return f"Please enter `{ctx.prefix}{ctx.invoked_with} {ctx.command.signature}`."


class AdminPermissionException(FancyContextException):
    async def formatMessage(self, ctx):
        usernick = ctx.author.display_name
        return f"{usernick} tried to run an admin command."

    async def formatUserMessage(self, ctx):
        usernick = ctx.author.display_name
        return f"{usernick} tried to run an admin command. This incident will be reported."


class MultilineAsNonFirstCommandException(FancyContextException):
    async def formatMessage(self, ctx):
        usernick = ctx.author.display_name
        return f"{usernick} tried to run a multi-line command in the middle of a sequence."

    async def formatUserMessage(self, ctx):
        return "You are unable to run a command that takes a multi-line argument in the middle of a batch command sequence. Please try running these commands seperately."


def patch():
    commands.errors.FancyException = FancyException
    commands.FancyException = FancyException
    commands.errors.FancyContextException = FancyContextException
    commands.FancyContextException = FancyContextException
    commands.errors.ArgumentException = ArgumentException
    commands.ArgumentException = ArgumentException
    commands.errors.AdminPermissionException = AdminPermissionException
    commands.AdminPermissionException = AdminPermissionException
    commands.errors.MultilineAsNonFirstCommandException = MultilineAsNonFirstCommandException
    commands.MultilineAsNonFirstCommandException = MultilineAsNonFirstCommandException



import sys
import logging
import traceback
from datetime import datetime

import discord.utils
from discord import Client
from discord.ext import commands

logger = logging.getLogger(__package__)


old_init = Client.__init__


def init(self, *args, username=None, activity=None, **kwargs):
    old_init(self, *args, **kwargs)

    self._init_username = username
    self._init_activity = activity


def oauth_url(self, *args, **kwargs):
    return discord.utils.oauth_url(self.user.id, *args, **kwargs)


old_activity_getter = Client.activity.fget


def activity_getter(self):
    if not self.guilds:
        return old_activity_getter(self)
    return self.guilds[0].me.activity


activity = property(fget=activity_getter, fset=Client.activity.fset)


old_dispatch = Client.dispatch


def dispatch(self, event_name, *args, **kwargs):
    new_event_name = None

    if event_name == "ready":
        has_readied = getattr(self, "_has_readied", False)
        if not has_readied:
            new_event_name = "first_ready"
            self._has_readied = True
        else:
            new_event_name = "reconnect_ready"

    old_dispatch(self, event_name, *args, **kwargs)
    if new_event_name:
        old_dispatch(self, new_event_name, *args, **kwargs)


async def on_first_ready(self):
    self._load_end = datetime.now()

    if self._init_username is not None and self.user.name != self._init_username:
        try:
            await self.user.edit(username=self._init_username)
        except discord.errors.HTTPException:
            logger.warn("We can't change the username this much!")

    if self._init_activity is not None:
        await self.change_presence(activity=self._init_activity)


def formatTraceback(err) -> str:
    return "".join(traceback.format_exception(type(err), err, err.__traceback__))


async def on_error(self, event, *args, **kwargs):
    """|coro|

    The default error handler provided by the client.

    By default this prints to :data:`sys.stderr` however it could be
    overridden to have a different implementation.
    Check :func:`~discord.on_error` for more details.
    """
    _, error, _ = sys.exc_info()
    # Get actual error
    err = getattr(error, "original", error)
    # DigiException handling
    if isinstance(err, commands.FancyException):
        message = err.formatMessage()
        if message is not None:
            logger.log(err.level, message)
            logger.error(formatTraceback(error))
    elif isinstance(err, commands.FancyContextException):
        message = str(err)
        if message is not None:
            logger.log(err.level, message)
            logger.error(formatTraceback(error))
    else:
        logger.error(f"Ignoring exception in {event}")
        logger.error(formatTraceback(error))


old_run = Client.run


def run(self, *args, **kwargs):
    self._load_start = datetime.now()
    result = old_run(self, *args, **kwargs)
    if hasattr(self, "on_disconnect"):
        self.on_disconnect()
    return result


def load_time_getter(self):
    start = getattr(self, "_load_start", None)
    end = getattr(self, "_load_end", None)
    if start is None or end is None:
        return None
    return end - start


load_time = property(fget=load_time_getter)


def patch():
    Client.__init__ = init
    Client.oauth_url = oauth_url
    Client.activity = activity
    Client.on_error = on_error
    Client.dispatch = dispatch
    Client.run = run
    Client.load_time = load_time
    Client.on_first_ready = on_first_ready



from discord import Member

# List of roles that map to each gender
gender_roles = {
    "m": ["male", "boy", "man", "masculine", "ftm"],
    "f": ["female", "girl", "woman", "feminine", "mtf"],
    "x": ["non-binary", "nonbinary"]
}
# Add "trans " prefix alterative for each role
for gender, roles in gender_roles.items():
    # Iterate over tuple(roles) so that we append to the same list we are iterating
    for role in tuple(roles):
        roles.append("trans " + role)

# List of roles that map to each pronoun
pronoun_roles = {
    "he": ["he/his", "he"],
    "she": ["she/her", "she"],
    "they": ["they/them", "they"]
}

# Map each gender to a pronoun
gender_to_pronoun = {
    "m": "he",
    "f": "she",
    "x": "they",
}

# Map each pronoun to a gender
pronoun_to_gender = {
    "he": "m",
    "she": "f",
    "they": "x"
}


@property
def rolelist(self):
    return [role.name for role in self.roles]


def get_genders(roles):
    genders = []
    for gender, roles_list in gender_roles.items():
        if any(role.name.lower() in roles_list for role in roles):
            genders.append(gender)
    return genders


def get_pronouns(roles):
    pronouns = []
    for pronoun, roles_list in pronoun_roles.items():
        if any(role.name.lower() in roles_list for role in roles):
            pronouns.append(pronoun)
    return pronouns


@property
def gender(self):
    genders = get_genders(self.roles)

    # One gender found
    if len(genders) == 1:
        gender = genders[0]
        # We don't support non-binary gender, yet.
        if gender == "x":
            gender = None
        return gender

    # Multiple genders found, result is ambigious
    if len(genders) > 1:
        return None

    # No genders found, fallback to pronouns
    pronouns = get_pronouns(self.roles)
    # One pronoun found
    if len(pronouns) == 1:
        gender = pronoun_to_gender[pronouns[0]]
        # We don't support non-binary gender, yet.
        if gender == "x":
            gender = None
        return gender

    # No genders found, and either ambigious or no pronouns found
    return None


@property
def pronouns(self):
    pronouns = get_pronouns(self.roles)

    # One pronoun found
    if len(pronouns) == 1:
        pronoun = pronouns[0]
        return pronoun

    # Multiple pronouns found, result is ambigious
    if len(pronouns) > 1:
        return None

    # No pronouns found, fallback to gender
    genders = get_genders(self.roles)
    # One gender found
    if len(genders) == 1:
        pronoun = gender_to_pronoun[genders[0]]
        return pronoun

    # No pronouns found, and either ambigious or no genders found
    return None


def patch():
    Member.rolelist = rolelist
    Member.gender = gender
    Member.pronouns = pronouns



from discord.ext.commands import Command, Cog
from discord.ext.commands.core import wrap_callback

old_init = Command.__init__
old_short_doc = Command.short_doc


def __init__(self, *args, category=None, multiline=False, **kwargs):
    self.category = category
    self.multiline = multiline
    old_init(self, *args, **kwargs)


async def dispatch_error(self, ctx, error):
    ctx.command_failed = True
    cog = self.cog
    try:
        command_on_error = self.on_error
    except AttributeError:
        command_on_error = None

    handled = False

    if command_on_error is not None:
        wrapped_command_on_error = wrap_callback(command_on_error)

        try:
            if cog is not None:
                await wrapped_command_on_error(cog, ctx, error)
            else:
                await wrapped_command_on_error(ctx, error)
            handled = True
        except Exception as e:
            error = e

    if not handled and cog is not None:
        cog_on_error = Cog._get_overridden_method(cog.cog_command_error)
        if cog_on_error is not None:
            wrapped_cog_on_error = wrap_callback(cog_on_error)
            try:
                await wrapped_cog_on_error(ctx, error)
                handled = True
            except Exception as e:
                error = e

    if not handled:
        ctx.bot.dispatch("command_error", ctx, error)


@property
def short_doc(self):
    return old_short_doc.fget(self) or "-"


@property
def alias_string(self):
    aliases = ""
    if self.aliases:
        aliases = "*(" + ", ".join(self.aliases) + ")*"
    return aliases


@property
def name_string(self):
    return f"**{self.name}**"


def patch():
    Command.__init__ = __init__
    Command.dispatch_error = dispatch_error
    Command.short_doc = short_doc
    Command.alias_string = alias_string
    Command.name_string = name_string



import asyncio

from discord.voice_client import VoiceClient


# Wrap the callback version in a function that returns a future
def play_until_done_future(voiceclient, source):
    loop = asyncio.get_running_loop()
    fut = loop.create_future()

    def after(err):
        if err:
            fut.set_exception(err)
            return
        fut.set_result(None)

    voiceclient.play(source, after=after)

    return fut


# Wrap the future version in a coroutine
async def play_until_done(self, source):
    return await play_until_done_future(self, source)


def patch():
    VoiceClient.play_until_done = play_until_done



from discord import Embed

old_add_field = Embed.add_field


def add_field(self, *, name=None, value=None, inline=True):
    if not name:
        name = "\u200b"
    if not value:
        value = "\u200b"

    return old_add_field(self, name=name, value=value, inline=inline)


def patch():
    Embed.add_field = add_field



from discord.abc import Messageable
import discord

old_history = Messageable.history


def wrapSnowflake(name, kwargs):
    if name not in kwargs:
        return
    value = kwargs[name]
    if isinstance(value, int):
        kwargs[name] = discord.Snowflake(value)


def history(self, **kwargs):
    wrapSnowflake("before", kwargs)
    wrapSnowflake("after", kwargs)
    wrapSnowflake("around", kwargs)

    return old_history(self, **kwargs)


def patch():
    Messageable.history = history



from copy import copy
import traceback
import logging

import discord.errors
from discord.ext.commands.bot import BotBase
from discord.ext.commands import Cog
from discord.ext import commands

logger = logging.getLogger(__package__)


old_init = BotBase.__init__


def init(self: BotBase, *args, extensions=(), **kwargs):
    old_init(self, *args, **kwargs)

    for e in extensions:
        self.load_extension(e)


class BadMultilineCommand(commands.errors.CommandError):
    """A multiline command is being run in the middle of a list of commands"""
    pass


def find_one(iterable):
    try:
        return next(iterable)
    except StopIteration:
        return None


async def process_commands(self, message):
    if message.author.bot:
        return

    # F*** smart quotes.
    message.content = message.content.replace("“", "\"")
    message.content = message.content.replace("”", "\"")
    message.content = message.content.replace("’", "'")
    message.content = message.content.replace("‘", "'")

    contexts = []

    ctx = await self.get_context(message)

    # No command found, invoke will handle it
    if not ctx.command:
        await self.invoke(ctx)
        return

    # One multiline command (command string starts with a multiline command)
    if ctx.command.multiline:
        await self.invoke(ctx)
        return

    # Multiple commands (first command is not multiline)
    lines = message.content.split("\n")
    messages = []
    for line in lines:
        msg = copy(message)
        msg.content = line
        messages.append(msg)
    contexts = [await self.get_context(msg) for msg in messages]

    # If at least one of the lines does not start with a prefix, then ignore all the lines
    not_command = find_one(ctx for ctx in contexts if ctx.invoked_with is None)
    if not_command:
        return

    # If at least one of the lines don't match to a command, then throw an error for that command
    bad_command = find_one(ctx for ctx in contexts if ctx.command is None)
    if bad_command:
        await self.invoke(bad_command)
        return

    # If at least one of the lines is a multi-line command (these are only allowed as the first command)
    multiline_command = find_one(ctx for ctx in contexts if ctx.command.multiline)
    if multiline_command:
        username = multiline_command.author.display_name
        await multiline_command.command.dispatch_error(multiline_command, commands.errors.BadMultilineCommand(f"{username} tried to run a multi-line command in the middle of a sequence."))
        return

    # If all the lines have a command, then run them in order
    for ctx in contexts:
        await self.invoke(ctx)


def formatTraceback(err) -> str:
    return "".join(traceback.format_exception(type(err), err, err.__traceback__))


class Emojis:
    info = ":information_source:"
    warning = ":warning:"
    error = ":no_entry:"


async def on_command_error(self, ctx, exception):
    """|coro|

    The default command error handler provided by the bot.

    By default this prints to :data:`sys.stderr` however it could be
    overridden to have a different implementation.

    This only fires if you do not specify any listeners for command error.
    """
    if self.extra_events.get('on_command_error', None):
        return

    if hasattr(ctx.command, 'on_error'):
        return

    cog = ctx.cog
    if cog:
        if Cog._get_overridden_method(cog.cog_command_error) is not None:
            return

    # Get actual error
    err = getattr(exception, "original", exception)

    emojis = on_command_error.emojis

    # If we got some bad arguments, use a generic argument exception error
    if isinstance(err, commands.BadUnionArgument) or isinstance(err, commands.MissingRequiredArgument) or isinstance(err, commands.BadArgument):
        err = commands.ArgumentException()

    if isinstance(err, commands.NotOwner):
        err = commands.AdminPermissionException()

    if isinstance(err, commands.BadMultilineCommand):
        err = commands.MultilineAsNonFirstCommandException()

    if isinstance(err, commands.FancyContextException):
        # DigiContextException handling
        message = await err.formatMessage(ctx)
        if message is not None:
            logger.log(err.level, message)
        userMessage = await err.formatUserMessage(ctx)
        if userMessage is not None:
            await ctx.send(f"{emojis.warning} {userMessage}")
    elif isinstance(err, commands.FancyException):
        # DigiException handling
        message = err.formatMessage()
        if message is not None:
            logger.log(err.level, message)
        userMessage = err.formatUserMessage()
        if userMessage is not None:
            await ctx.send(f"{emojis.warning} {userMessage}")
    elif isinstance(err, commands.CommandNotFound):
        pass
    elif isinstance(err, commands.MissingRequiredArgument):
        await ctx.send(f"{emojis.warning} Missing required argument(s) for `{ctx.prefix}{ctx.command}`.")
    elif isinstance(err, commands.ExpectedClosingQuoteError):
        await ctx.send(f"{emojis.warning} Mismatched quotes in command.")
    elif isinstance(err, commands.InvalidEndOfQuotedStringError):
        await ctx.send(f"{emojis.warning} No space after a quote in command. Are your arguments smushed together?")
    elif isinstance(err, commands.UnexpectedQuoteError):
        await ctx.send(f"{emojis.warning} Why is there a quote here? I'm confused...")
    elif isinstance(err, commands.CheckFailure):
        await ctx.send(f"{emojis.error} You do not have permission to run this command.")
    elif isinstance(err, commands.CommandOnCooldown):
        await ctx.send(f"{emojis.info} You're using that command too fast! Try again in a moment.")
    elif isinstance(err, discord.errors.ClientException):
        await ctx.send(f"{emojis.error} {err.args[0]}")
    else:
        # Default command error handling
        await ctx.send(f"{emojis.error} Something went wrong.")
        logger.error(f"Ignoring exception in command {ctx.command}:")
        logger.error(formatTraceback(exception))


on_command_error.emojis = Emojis


async def on_message_edit(self, before, after):
    # run editted commands
    if before.content == after.content:
        return
    await self.process_commands(after)


def patch():
    commands.errors.BadMultilineCommand = BadMultilineCommand
    commands.BadMultilineCommand = BadMultilineCommand
    BotBase.__init__ = init
    BotBase.process_commands = process_commands
    BotBase.on_command_error = on_command_error
    BotBase.on_message_edit = on_message_edit



import logging

import discord.ext.tasks

logger = logging.getLogger(__package__)


def safe_coro(func):
    async def wrapped(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except Exception as e:
            logger.error(e)
    return wrapped


def safe_loop(*args, **kwargs):
    def decorator(func):
        return discord.ext.tasks.loop(*args, **kwargs)(safe_coro(func))
    return decorator


def patch():
    discord.ext.tasks.safe_loop = safe_loop



__version__ = "1.2.2"

from . import bot, command, embed, member, client, messageable, snowflake, tasks, voiceclient, errors


def patch():
    embed.patch()
    command.patch()
    member.patch()
    bot.patch()
    client.patch()
    messageable.patch()
    snowflake.patch()
    tasks.patch()
    voiceclient.patch()
    errors.patch()
